import { Serializeable, Serializer } from "./serializeable";
import { MessageRecourceRecord } from "./types";
import { Label } from "./label";

export class RecourceRecord implements Partial<MessageRecourceRecord>, Serializeable {
   /**
    * This value can be set to identify if specific record is already set
    */
   public Identifier: string;

   private _NAME = new Label();
   public get NAME() { return this._NAME.name };
   public set NAME(name) { this._NAME.name = name };

   private _TYPE: number;
   public get TYPE() { return this._TYPE }
   public set TYPE(value) {
      if (value < 0 || value > 65535) throw new TypeError("TYPE Range: 0 - 65.535")
      this._TYPE = value;
   }

   private _CLASS: number = 1;
   public get CLASS() { return this._CLASS }
   public set CLASS(value) {
      if (value < 0 || value > 65535) throw new TypeError("CLASS Range: 0 - 65.535")
      this._CLASS = value;
   }

   private _TTL: number;
   public get TTL() { return this._TTL }
   public set TTL(value) {
      if (value < 0 || value > 4294967295) throw new TypeError("TTL Range: 0 - 4.294.967.295")
      this._TTL = value;
   }

   constructor(data?: Partial<MessageRecourceRecord>) {
      if (data) {
         for (let key in data) {
            this[key] = data[key];
         }
      }
   }

   protected dataLock?: () => Serializer;

   public lock() {
      if (!this.dataLock) return {
         length: 0,
         serialize: (buffer: Buffer, offset: number) => offset
      }

      let name = this._NAME.lock();
      let type = this.TYPE;
      let cl = this.CLASS;
      let ttl = this.TTL;
      let data = this.dataLock();
      let length = name.length + 2 + 2 + 4 + 2 + data.length;
      return {
         length: length,
         serialize: (buffer: Buffer, offset: number) => {
            offset = name.serialize(buffer, offset);
            buffer.writeUInt16BE(type, offset)
            offset += 2
            buffer.writeUInt16BE(cl, offset)
            offset += 2
            buffer.writeUInt32BE(ttl, offset)
            offset += 4
            buffer.writeUInt16BE(data.length, offset)
            offset += 2
            return data.serialize(buffer, offset);
         }
      }
   }
}

export class RDATARecord extends RecourceRecord {
   public RDATA: Buffer;

   get RDLENGTH() {
      return this.RDATA.length;
   }

   dataLock = () => {
      return {
         length: this.RDATA.length,
         serialize: (buffer: Buffer, offset: number) => {
            this.RDATA.copy(buffer, offset);
            return offset + this.RDATA.length;
         }
      }
   }
}